@@ -3,12 +3,10 @@ package Syntax::Feature::Try;
 use strict;
 use warnings;
 use XSLoader;
-use B::Hooks::OP::Check;
-use B::Hooks::OP::PPAddr;
 use Scalar::Util qw/ blessed /;
 
 BEGIN {
-    our $VERSION = '0.003';
+    our $VERSION = '0.004';
     XSLoader::load();
 }
 
@@ -20,21 +18,28 @@ sub uninstall {
     $^H{+HINTKEY_ENABLED} = 0;
 }
 
+# TODO convert "our" to "my" variables
+our $end_of_block;
+our $return_values;
+
 sub _statement {
     my ($try_block, $catch_list, $finally_block) = @_;
 
+    local $end_of_block;
+    my $return;
     local $@;
+    # TODO deduplicate try/catch/finally blocks code
     eval {
         BEGIN { $^H{+HINTKEY_BLOCK} = BLOCK_TRY }
-        $try_block->();
+        $return = run_block(\&$try_block);
     };
     my $exception = $@;
-    if ($exception) {
+    if ($exception and $catch_list) {
         my $handler = _get_exception_handler($exception, $catch_list);
         if ($handler) {
             eval {
                 BEGIN { $^H{+HINTKEY_BLOCK} = BLOCK_CATCH }
-                $handler->($exception);
+                $return = run_block(\&$handler, $exception);
             };
             $exception = $@;
         }
@@ -43,13 +48,16 @@ sub _statement {
     if ($finally_block) {
         {
             BEGIN { $^H{+HINTKEY_BLOCK} = BLOCK_FINALLY }
-            $finally_block->();
+            $return = run_block(\&$finally_block) || $return;
         }
     }
 
     if ($exception) {
         _rethrow($exception);
     }
+
+    $return_values = $return;
+    return $return;
 }
 
 sub _get_exception_handler {
@@ -77,6 +85,13 @@ sub _rethrow {
     die $exception;
 }
 
+sub _get_return_value {
+    my $return = $return_values;
+    undef $return_values;
+
+    return wantarray ? @$return : $return->[0];
+}
+
 1;
 
 __END__
@@ -171,11 +186,22 @@ It is always executed (even if try or catch block throw an error).
         ...
     }
     finally {
-        $fh->close;
+        $fh->close if $fh;
     }
 
 B<WARNING>: If finally block throws an exception,
 originaly thrown exception (from try/catch block) is discarded.
+You can convert errors inside finally block to warnings:
+
+    try {
+        # try block
+    }
+    finally {
+        try {
+            # cleanup code
+        }
+        catch ($e) { warn $e }
+    }
 
 =head1 Exception::Class
 
@@ -194,6 +220,26 @@ This module is compatible with Exception::Class
         # handle error here
     }
 
+=head2 return from subrutine
+
+This module supports also calling "return" inside try/catch/finally blocks
+to return values from subrutine.
+
+    sub read_config {
+        my $file;
+        try {
+            $fh = IO::File->new(...);
+            return $fh->getline; # it returns value from subrutine "read_config"
+        }
+        catch ($e) {
+            # log error
+        }
+        finally {
+            $fh->close();
+        }
+    }
+
+
 =head1 CAVEATS
 
 =head2 @_
@@ -201,26 +247,15 @@ This module is compatible with Exception::Class
 C<@_> is not accessible inside try/catch/finally blocks,
 because these blocks are internally called in different context.
 
-=head2 return, wantarray
-
-C<return> and C<wantarray> is not working inside try/catch/finally blocks,
-because these blocks are internally called in different context.
-
 =head2 next, last, redo
 
 C<next>, C<last> and C<redo> is not working inside try/catch/finally blocks,
 because these blocks are internally called in different context.
 
-=head1 TODO
-
-=over
-
-=item return, wantarray, ...
-
-=back
-
 =head1 BUGS
 
+None bugs known.
+
 =head1 SEE ALSO
 
 L<syntax> - Active syntax extensions
